using System;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using TaxControls;

public class Demo : Form
{
	private double      UNEMPLOYMENT_RATE = 0.05;     // normally, you'd put this in a configuration dialog
	private CTaxControl taxCtl;
	private ComboBox    lstPeriod;
	private TextBox     txtEarnings;
	private TextBox     txtYTDEarnings;
	private TextBox     txtExemptions;
	private TextBox     txtMiscellaneous;
	private CheckBox    cbMultipleJobs;
	private CheckBox    cbSupplemental;
	private CheckBox    cbRound;
	private ComboBox    lstFederalStatus;
	private ComboBox    lstState;
	private Label       lblLocal;
	private ComboBox    lstLocal;
	private Label       lblStateStatus;
	private ComboBox    lstStateStatus;
	private ToolTip     tooltip;
	private TableLayoutPanel     pnlMain;
	private TableLayoutPanel     pnlInput;
	private TableLayoutPanel     pnlResult;

	// These form labels change based on the state or local taxes:
	private Label       lblStateExemptions;
	private TextBox     txtStateExemptions;
	private Label       lblStateMiscellaneous;
	private TextBox     txtStateMiscellaneous;
	private Label       lblStateAuxiliary;
	private TextBox     txtStateAuxiliary;
	private Label       lblHours;            // used only for OR WBF
	private TextBox     txtHours;            // used only for OR WBF
	private TableLayoutPanel pnlState;       // parent for these fields

	private TableLayoutPanel pnlPaycheck;
	private TableLayoutPanel pnlEmployer;

	static public void Main ()
	{
    System.Console.WriteLine(System.IO.Directory.GetCurrentDirectory());
		Application.Run (new Demo());
	}

	public Demo()
	{
		// Initialize the TaxControl component.
		taxCtl = new CTaxControl();

		// DataFilename points to the tax file (demo.tax if from last year).
		taxCtl.DataFilename = "../../demo.tax";
		taxCtl.UnemploymentRate = UNEMPLOYMENT_RATE;

		// Data is not loaded until actually used.  To force loading now, do:
		// taxCtl.Refresh();
		
		LayoutForm();
	}

	private void LayoutForm()
	{
		Width = 630;
		Height = 580;
		try{
			Icon = new Icon("taxcontrols.ico");
		} catch (Exception) {
		//	Console.WriteLine(e.Message);
		}
		tooltip = new ToolTip();
		Text = "TaxControls Demo";
		pnlMain = new TableLayoutPanel();
		pnlMain.ColumnCount = 2;
		pnlMain.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 50F));
		pnlMain.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 50F));
		pnlMain.Dock = DockStyle.Fill;
		Controls.Add(pnlMain);

		pnlInput = new TableLayoutPanel();
		pnlInput.ColumnCount = 1;
		pnlInput.Dock = DockStyle.Fill;
		pnlInput.AutoSize = true;
		pnlMain.Controls.Add(pnlInput);

		LayoutGeneral(pnlInput);
		LayoutFederal(pnlInput);
		LayoutStateLocal(pnlInput);

		pnlResult = new TableLayoutPanel();
		pnlResult.ColumnCount = 1;
		pnlResult.Dock = DockStyle.Fill;
		pnlResult.AutoSize = true;
		pnlMain.Controls.Add(pnlResult);

		LayoutPaycheck(pnlResult);
		LayoutEmployer(pnlResult);

		LayoutButtons();
		Compute_Click(null, null);
	}

	private GroupBox MakeGB(string title, Control par)
	{
		GroupBox g = new GroupBox();
		g.Text = title;
		g.Font = new Font(g.Font, FontStyle.Bold);
		g.Dock = DockStyle.Top;
		g.AutoSize = true;
		par.Controls.Add(g);
		return g;
	}

	private void LayoutGeneral(TableLayoutPanel par)
	{
		GroupBox g = MakeGB("General", par);

		TableLayoutPanel pnl = new TableLayoutPanel();
		pnl.Dock = DockStyle.Top;
		pnl.AutoSize = true;
		pnl.Font = new Font(pnl.Font, FontStyle.Regular);
		g.Controls.Add(pnl);
		pnl.ColumnCount = 2;

		AddPayPeriodChoice(pnl);
                cbSupplemental = AddCheckBox("Supplemental:", "Check for separate supplemental (bonus) payment.", pnl);
		txtEarnings = AddField("Earnings:", "2000.00", "Gross wages for the current pay period", pnl);
		txtYTDEarnings = AddField("YTD Earnings:", "0.00", "Gross wages this year BEFORE the current pay period", pnl);
		cbRound = AddCheckBox("Round when optional:", "When checked, all roundable taxes will be rounded", pnl);
		cbMultipleJobs = AddCheckBox( "W-4 Step 2(c) is checked:", "Indicates multiple jobs or spouse works", pnl);
	}

	private void LayoutFederal(TableLayoutPanel par)
	{
		GroupBox g = MakeGB("Federal", par);

		TableLayoutPanel pnl = new TableLayoutPanel();
		pnl.ColumnCount = 2;
		pnl.Dock = DockStyle.Top;
		pnl.AutoSize = true;
		pnl.Font = new Font(pnl.Font, FontStyle.Regular);
		g.Controls.Add(pnl);

		// We use "FIT" if the employee has NOT signed the 2020 (or later) W-4.
		// Here we assume they have signed, so we use "Federal Income Tax".
		// For specific instructions on computing this tax, see
		//       https://www.taxupdate.com/federal.xls
		//
		// Federal Income Tax:
                string tip = "Federal Filing Status";
		MakeLabel("Filing Status:", tip, pnl);
		lstFederalStatus = FilingStatusChoice(tip, pnl);
		txtExemptions = AddField("Exemptions:", "2000", "Federal W-4 line 3 amount.", pnl);
		
		// Because there's so much variance in formulas, we must have some general-purpose
		// variables: Miscellaneous and Auxiliary.
		// Here is how we get labels and instructions for Miscellaneous & Auxiliary:
		Tax fit = taxCtl.Taxes.Item("Federal Income Tax");
		string miscLabel = fit.MiscellaneousTag;
		if (miscLabel != "")
		{
			txtMiscellaneous = AddField(miscLabel + ":", "0.00", fit.MiscellaneousInstructions, pnl);
		}

	}

	private void LayoutStateLocal(TableLayoutPanel par)
	{
		GroupBox g = MakeGB("State and Local", par);

		pnlState = new TableLayoutPanel();
		pnlState.ColumnCount = 2;
		pnlState.Dock = DockStyle.Top;
		pnlState.Height = 140;
		pnlState.Font = new Font(pnlState.Font, FontStyle.Regular);
		g.Controls.Add(pnlState);

		// These labels have variable Text values, based on the state selected:
		lblStateExemptions = MakeVariableLabel(par);
		txtStateExemptions = AddField("", "", "", par);
		RemoveOptionalRow(lblStateExemptions, txtStateExemptions, par);

		lblStateMiscellaneous = MakeVariableLabel(par);
		txtStateMiscellaneous = AddField("", "", "", par);
		RemoveOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, par);

		lblStateAuxiliary = MakeVariableLabel(par);
		txtStateAuxiliary = AddField("", "", "", par);
		RemoveOptionalRow(lblStateAuxiliary, txtStateAuxiliary, par);

		AddStateChoice(pnlState);
                string tip = "State Filing Status";
		lblStateStatus = MakeLabel("Filing Status:", tip, pnlState);
		lstStateStatus = FilingStatusChoice(tip, pnlState);
		RemoveOptionalRow(lblStateStatus, lstStateStatus, pnlState);

		AddLocalChoice(pnlState);

		// for OR WBF & WA LNI:
		lblHours = MakeLabel("Hours worked", "Number of hours worked in this pay period", pnlState);
		txtHours = new TextBox();
		txtHours.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		txtHours.TextAlign = HorizontalAlignment.Right;
		tooltip.SetToolTip(txtHours, "Number of hours worked in this pay period");
		AddOptionalRow(lblHours, txtHours, null, pnlState);
		RemoveOptionalRow(lblHours, txtHours, pnlState);

		State_Change(null, null);   // default state is AL, so show the appropriate fields
	}

	private void LayoutPaycheck(TableLayoutPanel par)
	{
		GroupBox g = MakeGB("Paycheck", par);
		pnlPaycheck = new TableLayoutPanel();
		pnlPaycheck.Dock = DockStyle.Top;
		pnlPaycheck.Font = new Font(pnlPaycheck.Font, FontStyle.Regular);
		pnlPaycheck.ColumnCount = 2;
		pnlPaycheck.Height = 200;
		g.Controls.Add(pnlPaycheck);
	}

	private void LayoutEmployer(TableLayoutPanel par)
	{
		GroupBox g = MakeGB("Employer Taxes", par);
		pnlEmployer = new TableLayoutPanel();
		pnlEmployer.Dock = DockStyle.Top;
		pnlEmployer.Font = new Font(pnlEmployer.Font, FontStyle.Regular);
		pnlEmployer.ColumnCount = 2;
		pnlEmployer.Height = 100;
		g.Controls.Add(pnlEmployer);
	}

	class State
	{
		public string abbrev { get; set; }
		public string name { get; set; }
		public static State New(string a, string nm) {
			State s = new State();
			s.abbrev = a;
			s.name = nm;
			return s;
		}
	}

	State[] states = new State[] {
		State.New("", "<select>")
			,State.New("AK", "Alaska")
			,State.New("AL", "Alabama")
			,State.New("AZ", "Arizona")
			,State.New("AR", "Arkansas")
			,State.New("CA", "California")
			,State.New("CO", "Colorado")
			,State.New("CT", "Connecticut")
			,State.New("DE", "Delaware")
			,State.New("DC", "DC")  // full name too wide
			,State.New("FL", "Florida")
			,State.New("GA", "Georgia")
			,State.New("HI", "Hawaii")
			,State.New("ID", "Idaho")
			,State.New("IL", "Illinois")
			,State.New("IN", "Indiana")
			,State.New("IA", "Iowa")
			,State.New("KS", "Kansas")
			,State.New("KY", "Kentucky")
			,State.New("LA", "Louisiana")
			,State.New("ME", "Maine")
			,State.New("MD", "Maryland")
			,State.New("MA", "Massachusetts")
			,State.New("MI", "Michigan")
			,State.New("MN", "Minnesota")
			,State.New("MS", "Mississippi")
			,State.New("MO", "Missouri")
			,State.New("MT", "Montana")
			,State.New("NE", "Nebraska")
			,State.New("NH", "New Hampshire")
			,State.New("NJ", "New Jersey")
			,State.New("NM", "New Mexico")
			,State.New("NY", "New York")
			,State.New("NV", "Nevada")
			,State.New("NC", "North Carolina")
			,State.New("ND", "North Dakota")
			,State.New("OH", "Ohio")
			,State.New("OK", "Oklahoma")
			,State.New("OR", "Oregon")
			,State.New("PA", "Pennsylvania")
			,State.New("RI", "Rhode Island")
			,State.New("SC", "South Carolina")
			,State.New("SD", "South Dakota")
			,State.New("TN", "Tennessee")
			,State.New("TX", "Texas")
			,State.New("UT", "Utah")
			,State.New("VT", "Vermont")
			,State.New("VA", "Virginia")
			,State.New("WA", "Washington")
			,State.New("WV", "West Virginia")
			,State.New("WI", "Wisconsin")
			,State.New("WY", "Wyoming")
	};

	private void AddStateChoice(Control par)
	{
		MakeLabel("State:", "Select state in which workplace resides.", par);
		lstState = new ComboBox();
		lstState.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		foreach(State st in states)
		{
			lstState.Items.Add(st.name);
		}
		par.Controls.Add(lstState);
		lstState.DropDownStyle = ComboBoxStyle.DropDownList;
		lstState.SelectedIndex = 2;    // Alabama
		lstState.SelectedIndexChanged += new EventHandler(State_Change);
	}

	private void AddLocalChoice(Control par)
	{
		// When we've selected a state with local taxes, we'll load this combobox:
		lblLocal = MakeLabel("Local:", "Select local jurisdiction of workplace.", par);
		lstLocal = new ComboBox();
		lstLocal.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		lstLocal.Items.Add("n/a");
		lstLocal.Enabled = false;
		lstLocal.DropDownStyle = ComboBoxStyle.DropDownList;
		lstLocal.SelectedIndexChanged += new EventHandler(Local_Change);
		par.Controls.Add(lstLocal);

		// Don't show it until a state with local taxes is selected:
		RemoveOptionalRow(lblLocal, lstLocal, par);
	}

	private void AddOptionalRow(Label lbl, Control ctl, string tip, Control par)
	{
		if (lbl.Parent != par)
		{
			par.Controls.Add(lbl);
			par.Controls.Add(ctl);
			if (tip != null){
				tooltip.SetToolTip(lbl, tip);
				tooltip.SetToolTip(ctl, tip);
			}
		}
	}

	private void RemoveOptionalRow(Label lbl, Control ctl, Control par)
	{
		if (lbl.Parent == par)
		{
			par.Controls.Remove(lbl);
			par.Controls.Remove(ctl);
		}
	}

	/*
	 * Depending on the state, we want to show/hide other fields.  We also adjust the labels and
	 * tooltips on those fields to match the terminology of the tax.
	 */
	private void SetStateFields()
	{
		pnlState.SuspendLayout();
		RemoveOptionalRow(lblStateExemptions, txtStateExemptions, pnlState);
		RemoveOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, pnlState);
		RemoveOptionalRow(lblStateAuxiliary, txtStateAuxiliary, pnlState);
		RemoveOptionalRow(lblStateStatus, lstStateStatus, pnlState);
		RemoveOptionalRow(lblLocal, lstLocal, pnlState);
		RemoveOptionalRow(lblHours, txtHours, pnlState);

		if (lstState.SelectedIndex > 0)
		{
			State st = states[lstState.SelectedIndex];
			try{
				Tax tax = taxCtl.Taxes.Item(st.abbrev);
				if (!tax.Discontinued)
				{
					// State Exemptions may be called something else, so get the name right:
					string exemptTag = tax.StateExemptionTag;

					// put the exemption instructions in a tooltip:
					string exemptInst = tax.StateExemptionInstructions;
					lblStateExemptions.Text = (exemptTag == "") ? "Exemptions: " : exemptTag + ":";

					// and show the fields:
					AddOptionalRow(lblStateStatus, lstStateStatus, null, pnlState);
					AddOptionalRow(lblStateExemptions, txtStateExemptions, exemptInst, pnlState);

					// Now do the same for Miscellaneous, if needed:
					string miscTag = tax.MiscellaneousTag;

					if (miscTag == "")
					{
						RemoveOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, pnlState);
					} else {
						string miscInst = tax.MiscellaneousInstructions;
						AddOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, miscInst, pnlState);
						lblStateMiscellaneous.Text = miscTag + ":";
					}

					// And for Auxiliary, if needed:
					string auxTag = tax.AuxiliaryTag;

					if (auxTag == "")
					{
						RemoveOptionalRow(lblStateAuxiliary, txtStateAuxiliary, pnlState);
					} else {
						string auxInst = tax.AuxiliaryInstructions;
						AddOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, auxInst, pnlState);
						lblStateAuxiliary.Text = auxTag + ":";
					}

					// if any local taxes, show the selector:
					if (HasLocals(st.abbrev)){
						AddOptionalRow(lblLocal, lstLocal, null, pnlState);
					}

					// Oregon's WBF needs hours worked:
					if (st.abbrev == "OR")
					{
						AddOptionalRow(lblHours, txtHours, null, pnlState);
					}
				} else if (tax.Name == "WA") {  // no state tax, but LNI is withheld
					Tax lni = taxCtl.Taxes.Item("WA LNI");
					string tip = $"WA LNI {lni.MiscellaneousInstructions}";
					lblStateMiscellaneous.Text = $"WA LNI {lni.MiscellaneousTag}";
					AddOptionalRow(lblStateMiscellaneous, txtStateMiscellaneous, tip, pnlState);
					AddOptionalRow(lblHours, txtHours, null, pnlState);
				}
			} catch (TaxControlException e) {
				Console.WriteLine(e.Message);
			}
		}
		pnlState.ResumeLayout(true);
	}

	private void SetLocalFields()
	{
		lstLocal.Items.Clear();
		if (lstState.SelectedIndex > 0)
		{
			State st = states[lstState.SelectedIndex];
			lstLocal.Items.Add("<select>");
			lstLocal.Enabled = true;
			lstLocal.SelectedIndex = 0;
			foreach(Tax t in taxCtl.Taxes)
			{
				// local tax names start with state abbrev
				// TaxTypes: "F" - federal
				//           "S" - state
				//           "L" - local
				//           "O" - other (unemployment, disability, ...)
				//           "P" - private (used in tax formulas)
				if (t.Name.StartsWith(st.abbrev) && t.TaxType == "L" && t.Withheld)  // don't include employer-paid taxes
				{
					lstLocal.Items.Add(t.Name);
				}
			}
		}
	}

	private bool HasLocals(string state_abbrev)
	{
		foreach(Tax t in taxCtl.Taxes)
		{
			if (t.Name.StartsWith(state_abbrev) && t.TaxType == "L" && t.Withheld)
			{
				return true;
			}
		}
		return false;
	}

	// if a local tax needs Miscellaneous or Auxiliary info, we need
	// to provide an appropriate form entry:
	private void SetLocalOptionalFields()
	{
		if (lstLocal.SelectedIndex > 0)
		{
			try{
				string name = (string)lstLocal.SelectedItem;
				Tax t = taxCtl.Taxes.Item(name);
				if (t.MiscellaneousTag != "")
				{
					Console.WriteLine("tag: {0}", t.MiscellaneousTag);
				}
				if (t.AuxiliaryTag != "")
				{
					Console.WriteLine("auxtag: {0}", t.AuxiliaryTag);
				}
			} catch (TaxControlException e) {
				Console.WriteLine(e.Message);
			}
		}

	}

	class PayPeriod
	{
		public int payPeriodsPerYear { get; set; }
		public string Text { get; set; }
	}

	private void AddPayPeriodChoice(Control par)
	{
		string tip =  "Current pay period length";
		MakeLabel("Pay Period:", tip, par);
		lstPeriod = new ComboBox();
		lstPeriod.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		lstPeriod.DataSource = new PayPeriod[] {
		    new PayPeriod { payPeriodsPerYear =   1, Text = "Yearly" },
		    new PayPeriod { payPeriodsPerYear =   4, Text = "Quarterly" },
		    new PayPeriod { payPeriodsPerYear =  12, Text = "Monthly" },
		    new PayPeriod { payPeriodsPerYear =  24, Text = "Bimonthly" },
		    new PayPeriod { payPeriodsPerYear =  26, Text = "Biweekly" },
		    new PayPeriod { payPeriodsPerYear =  52, Text = "Weekly" },
		    new PayPeriod { payPeriodsPerYear = 230, Text = "Daily" }
		};
		lstPeriod.DropDownStyle = ComboBoxStyle.DropDownList;
		par.Controls.Add(lstPeriod);
		lstPeriod.DisplayMember = "Text";
		lstPeriod.SelectedIndex = 2;   // Monthly is the default
		tooltip.SetToolTip(lstPeriod, tip);

		lstPeriod.DropDownStyle = ComboBoxStyle.DropDownList;
		lstPeriod.DisplayMember = "Text";
		lstPeriod.SelectedIndex = 2;   // Monthly is the default
		tooltip.SetToolTip(lstPeriod, tip);
	}

	private ComboBox FilingStatusChoice(string tip, Control par)
	{
		ComboBox cbox = new ComboBox();
		cbox.Items.Add("Single");                    // FilingStatus = 0
		cbox.Items.Add("Widow(er)");                 // FilingStatus = 1
		cbox.Items.Add("Head of Household");         // FilingStatus = 2
		cbox.Items.Add("Married Filing Jointly");    // FilingStatus = 3
		cbox.Items.Add("Married Filing Separately"); // FilingStatus = 4
		cbox.Items.Add("Married Both Employed");     // FilingStatus = 5
		
		cbox.Parent = par;
		cbox.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		cbox.DropDownStyle = ComboBoxStyle.DropDownList;
		cbox.SelectedIndex = 0;
		tooltip.SetToolTip(cbox, tip);
		return cbox;
	}

	private CheckBox AddCheckBox(string label, string tip, Panel par)
	{
		MakeLabel(label, tip, par);
		CheckBox cbox = new CheckBox();
		cbox.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		par.Controls.Add(cbox);
		tooltip.SetToolTip(cbox, tip);
		return cbox;
	}

	private TextBox AddField(string label, string val, string tip, Panel par)
	{
		MakeLabel(label, tip, par);

		TextBox tbox = new TextBox();
		tbox.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		tbox.Parent = par;
		tbox.TextAlign = HorizontalAlignment.Right;
		tbox.Text = val;
		tooltip.SetToolTip(tbox, tip);
		return tbox;
	}

	private void LayoutButtons()
	{
		Button btnHelp = new Button();
		Button btnCompute = new Button();
		Button btnExit = new Button();

		Panel panel = new Panel();
		panel.Height = btnCompute.Height + 2;
		panel.Dock = DockStyle.Bottom;
		Controls.Add(panel);

		btnHelp.Text = "&Help";
		btnHelp.Click += new EventHandler(Help_Click);
		btnHelp.Dock = DockStyle.Right;
		panel.Controls.Add(btnHelp);

		btnCompute.Text = "&Compute";
		btnCompute.Click += new EventHandler(Compute_Click);
		btnCompute.Dock = DockStyle.Right;
		panel.Controls.Add(btnCompute);

		btnExit.Text = "E&xit";
		btnExit.Click += new EventHandler(Exit_Click);
		btnExit.Dock = DockStyle.Right;
		panel.Controls.Add(btnExit);

		AcceptButton = btnCompute;
		CancelButton = btnExit;
	}

	private Label MakeLabel(string text, string tip, Control par)
	{
		Label lbl = new Label();
		lbl.Parent = par;
		lbl.Text = text;
		lbl.Anchor = AnchorStyles.Left | AnchorStyles.Top;
		lbl.AutoSize = true;
		tooltip.SetToolTip(lbl, tip);
		return lbl;
	}

	private Label MakeVariableLabel(Control par)
	{
		Label lbl = MakeLabel("", "", par);
		lbl.AutoSize = true;
		return lbl;
	}

	private Label MakeResultLabel(string text, string tip, Control par)
	{
		MakeLabel(text, tip, par);
		Label lbl = MakeLabel("0.00", tip, par);
		lbl.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		return lbl;
	}

	private void MakeResult(string text, string tip, double val, Control par, bool discontinued = false)
	{
		Label lbl1 = MakeLabel(text, tip, par);
		Label lbl2 = MakeLabel(val.ToString("0.00"), tip, par);
		lbl2.Anchor = AnchorStyles.Right | AnchorStyles.Top;
		if (discontinued){
			lbl1.ForeColor = Color.Gray;
			lbl2.ForeColor = Color.Gray;
		}
	}

	private void SkipLine(Control par)
	{
		Label lbl = new Label();
		lbl.Parent = par;
		lbl = new Label();
		lbl.Parent = par;
	}

	private void Help_Click (object sender, EventArgs e)
	{
		string os = System.Environment.OSVersion.ToString();
		if (os.StartsWith("Unix")){
			System.Diagnostics.ProcessStartInfo startInfo = new System.Diagnostics.ProcessStartInfo();
			startInfo.FileName = "xchm";    // another option is "chmsee"
			startInfo.Arguments = " TaxControls.chm";
			System.Diagnostics.Process.Start(startInfo);
		} else {
			Help.ShowHelp(this, "TaxControls.chm");
		}
	}

	private void Compute_Click (object sender, EventArgs e)
	{
		PayPeriod p = (PayPeriod)lstPeriod.SelectedValue;
		int stateNdx = lstState.SelectedIndex;
		taxCtl.Supplemental = cbSupplemental.Checked;

		// federal income tax:
		taxCtl.SelectedTax = "Federal Income Tax";          
		taxCtl.PayPeriodsPerYear = p.payPeriodsPerYear;
		taxCtl.Earnings = SafeParse(txtEarnings);
		taxCtl.Exemptions = (int) SafeParse(txtExemptions);
		taxCtl.YTDEarnings = SafeParse(txtYTDEarnings);
		taxCtl.Exemptions = (int) SafeParse(txtExemptions);
		taxCtl.FilingStatus = (FilingStatus) lstFederalStatus.SelectedIndex;
		taxCtl.RoundWhenOptional = cbRound.Checked;
		taxCtl.Miscellaneous = SafeParse(txtMiscellaneous);
		taxCtl.Auxiliary = cbMultipleJobs.Checked ? -1 : 0;

		double fit = RoundCurrency(taxCtl.TaxAmount());
		double ss = FindTax("FICA SS");
		double med = FindTax("FICA Med");
		double net = taxCtl.Earnings - fit - ss - med;

		// Paycheck results:
		pnlPaycheck.SuspendLayout();
		pnlPaycheck.Controls.Clear();
		MakeResult("Earnings", "Gross Earnings", taxCtl.Earnings, pnlPaycheck);
		SkipLine(pnlPaycheck);
		MakeResult("Federal Tax", "Federal Income Tax", fit, pnlPaycheck);
		MakeResult("Social Security", "Social Security (Employee portion)", ss, pnlPaycheck);
		MakeResult("Medicare", "Medicare (Employee portion)", med, pnlPaycheck);
		SkipLine(pnlPaycheck);

		// state:
		if (stateNdx != 0)
		{
			State st = states[stateNdx];
			taxCtl.SelectedTax = st.abbrev;          
			taxCtl.StateExemptions = (int)SafeParse(txtStateExemptions);
			taxCtl.FilingStatus = (FilingStatus) lstStateStatus.SelectedIndex;
			taxCtl.Miscellaneous = SafeParse(txtStateMiscellaneous);
			taxCtl.Auxiliary = SafeParse(txtStateAuxiliary);
			double stamt = RoundCurrency(taxCtl.TaxAmount());

			net -= stamt;
			Tax t_state = taxCtl.Taxes.Item(taxCtl.SelectedTax);
			if (t_state.Discontinued)
			{
				MakeResult(st.abbrev, "No state tax", stamt, pnlPaycheck, true);
			} else {
				MakeResult(st.abbrev, t_state.Title, stamt, pnlPaycheck);
			}
			net -= FindMisc(st.abbrev, true, pnlPaycheck);

			// local:
			if (HasLocals(st.abbrev) && lstLocal.SelectedIndex > 0)
			{
				taxCtl.SelectedTax = lstLocal.Text;
				double loc = RoundCurrency(taxCtl.TaxAmount());
				net -= loc;
				Tax t = taxCtl.Taxes.Item(taxCtl.SelectedTax);
				MakeResult(lstLocal.Text, t.Title, loc, pnlPaycheck);
			}
		}
		MakeResult("Net", "Net Pay", net, pnlPaycheck);
		pnlPaycheck.ResumeLayout(true);

		// Employer results:
		pnlEmployer.SuspendLayout();
		pnlEmployer.Controls.Clear();
		double ss_er = FindTax("FICA SS Employer");
		double med_er = FindTax("FICA Med Employer");
		double emp_tot = ss_er + med_er;
		MakeResult("Social Security", "Social Security (Employer)", ss_er, pnlEmployer);
		MakeResult("Medicare", "Medicare (Employee portion)", med_er, pnlEmployer);
		if (stateNdx != 0)
		{
			State st = states[stateNdx];
			emp_tot += FindMisc(st.abbrev, false, pnlEmployer);

			// There are a few local employer-paid taxes:
			if (lstLocal.SelectedIndex > 0)
			{
				string nm1 = $"{lstLocal.Text} Employer";
				if (taxCtl.HasTax(nm1))  // there usually isn't one
				{
					double loc_er = FindTax(nm1);
					Tax t1 = taxCtl.Taxes.Item(nm1);
					MakeResult(nm1, t1.Title, loc_er, pnlEmployer);
					emp_tot += loc_er;
				}
			}
		}

		MakeResult("Total", "Total Employer Taxes", emp_tot, pnlEmployer);
		pnlEmployer.ResumeLayout(true);

	}

	private double FindMisc(string abbr, bool withheld, Control par)
	{
		double ret = 0;
		double save = taxCtl.Auxiliary;
		foreach(Tax t in taxCtl.Taxes)
		{
			if (t.Name.StartsWith(abbr) && t.TaxType == "O" && t.Withheld == withheld)
			{
				if (t.Name.StartsWith("OR WBF") | t.Name.StartsWith("WA LNI"))
				{
					taxCtl.Auxiliary = SafeParse(txtHours);
				}
				taxCtl.SelectedTax = t.Name;
				double amt = taxCtl.TaxAmount();
				MakeResult(t.Name, t.Title, amt, par);
				ret += amt;
			}
		}
		taxCtl.Auxiliary = save;
		return ret;
	}

	private double FindTax(string nm)
	{
		double ret = 0;
		try {
			taxCtl.SelectedTax = nm;
			ret = RoundCurrency(taxCtl.TaxAmount());
		} catch (TaxControlException) {
		}
		return ret;
	}

	private double SafeParse(TextBox fld)
	{
		double res = 0;
		try {
			res = double.Parse(fld.Text);
		} catch (Exception) {
		}
		return res;
	}

	private bool RunningOnMono()
	{
		return Type.GetType("Mono.Runtime") != null;
	}

	private void State_Change (object sender, EventArgs e)
	{
		SetStateFields();
		SetLocalFields();
	}

	private void Local_Change (object sender, EventArgs e)
	{
		SetLocalOptionalFields();
	}

	private void Exit_Click (object sender, EventArgs e)
	{
		Close();
	}

  private double RoundCurrency(double amt)
  {
    return Math.Round(amt * 100) / 100;
  }
}

